#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <vector>
#include <algorithm>
#include "glwindow.h"

struct point {
	double x, y;
	double px, py;
	double ax, ay;
};
struct measurement {
	int p1, p2;
	double d;
	double stress;
	double alpha;
};
struct angle {
	int p1, p2, p3;
	double rad;
	double stress;
	double alpha;
};
std::vector<point> points;
std::vector<measurement> measurements;
std::vector<angle> angles;

void reload()
{
	FILE *file = fopen("spec.txt", "r");
	unsigned num_points;
	fscanf(file, "%u", &num_points);

	if (num_points < points.size()) {
		points.erase(points.begin() + num_points, points.end());
	} else if (num_points > points.size()) {
		for (int i = points.size(); i < num_points; ++i) {
			point p;
			if (i == 0) {
				p.x = p.y = 0.0;
			} else {
				p.x = rand() / (RAND_MAX + 1.0);
				p.y = rand() / (RAND_MAX + 1.0);
			}
			if (i == 1) {
				p.y = 0.0;
			}
			p.px = p.x;
			p.py = p.y;
			points.push_back(p);
		}
	}

	measurements.erase(measurements.begin(), measurements.end());
	angles.erase(angles.begin(), angles.end());

	for ( ;; ) {
		char buf[256];
		if (fscanf(file, "%s", &buf) != 1)
			break;

		if (buf[0] == 'd') {
			measurement l;
			if (fscanf(file, "%u %u %lf %lf", &l.p1, &l.p2, &l.d, &l.alpha) != 4)
				break;
			measurements.push_back(l);
		}
		if (buf[0] == 'a') {
			angle a;
			if (fscanf(file, "%u %u %u %lf %lf", &a.p1, &a.p2, &a.p3, &a.rad, &a.alpha) != 5)
				break;
			a.rad *= M_PI / 180.0;
			angles.push_back(a);
		}
	}

	fclose(file);
}

void recalc()
{
	double dt = 1.0/60.0;
	double k = 0.01;
	double k_angle = 0.02;
	double f = 0.02 * dt;

	for (unsigned i = 0; i < points.size(); ++i) {
		points[i].ax = points[i].ay = 0.0;
	}

	for (unsigned i = 0; i < measurements.size(); ++i) {
		double dx = points[measurements[i].p2].x - points[measurements[i].p1].x;	
		double dy = points[measurements[i].p2].y - points[measurements[i].p1].y;	
		double d = sqrt(dx*dx + dy*dy);
		measurements[i].stress = fabs(d - measurements[i].d);

		points[measurements[i].p1].ax += (d - measurements[i].d) * dx/d * k;	
		points[measurements[i].p1].ay += (d - measurements[i].d) * dy/d * k;	
		points[measurements[i].p2].ax -= (d - measurements[i].d) * dx/d * k;	
		points[measurements[i].p2].ay -= (d - measurements[i].d) * dy/d * k;
	}
	
	for (unsigned i = 0; i < angles.size(); ++i) {
		double cx = points[angles[i].p1].x - points[angles[i].p2].x;
		double cy = points[angles[i].p1].y - points[angles[i].p2].y;
		double c = sqrt(cx*cx + cy*cy);
		
		double bx = points[angles[i].p3].x - points[angles[i].p2].x;
		double by = points[angles[i].p3].y - points[angles[i].p2].y;
		double b = sqrt(bx*bx + by*by);

		double alpha = acos((bx*cx + by*cy) / (b*c));
		angles[i].stress = fabs(alpha - angles[i].rad);

		double tp3x = -by / b;
		double tp3y = bx / b;
		
		double tp1x = cy / c;
		double tp1y = -cx / c;

		points[angles[i].p1].ax -= (alpha - angles[i].rad) * tp1x * k_angle;	
		points[angles[i].p1].ay -= (alpha - angles[i].rad) * tp1y * k_angle;	
		points[angles[i].p3].ax -= (alpha - angles[i].rad) * tp3x * k_angle;	
		points[angles[i].p3].ay -= (alpha - angles[i].rad) * tp3y * k_angle;
	}

	// constraints
	if (points.size() >= 1) {
		points[0].ax = points[0].ay = 0.0;
	}
	if (points.size() >= 2) {
		points[1].ay = 0.0;
	}

	// Verlet integration
	for (unsigned i = 0; i < points.size(); ++i) {
		double nx = (2.0 - f) * points[i].x - (1.0 - f) * points[i].px + points[i].ax * dt * dt; 
		double ny = (2.0 - f) * points[i].y - (1.0 - f) * points[i].py + points[i].ay * dt * dt; 
		points[i].px = points[i].x;
		points[i].py = points[i].y;
		points[i].x = nx;
		points[i].y = ny;
	}
	
	if (points.size() >= 2) {
		if (points[1].x < 0.0) {
			points[1].x = -points[1].x;
			points[1].px = -points[1].px;
			points[1].ax = -points[1].ax;
		}
	}
}

void redraw(GLWindow &win)
{
	glClear(GL_COLOR_BUFFER_BIT);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	//glEnable(GL_LINE_SMOOTH);
	//glLineWidth(5.0);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.0, 1.0, -1.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glScalef(0.1, 0.1, 0.1);

	glPointSize(5.0);
	glColor4f(1.0, 1.0, 1.0, 1.0);

	glBegin(GL_POINTS);
	for (int i = 0; i < points.size(); ++i) {
		glVertex2f(points[i].x, points[i].y);
	}
	glEnd();

	glBegin(GL_LINES);
	for (int i = 0; i < measurements.size(); ++i) {
		double sfac = std::min(5.0 * (fabs(measurements[i].stress) / measurements[i].d), 1.0);
		double r, g, b;
		if (sfac < 0.5) {
			r = 2.0 * sfac;
			g = 1.0;
			b = 0.0;
		} else {
			r = 1.0;
			g = 1.0 - 2.0 * (sfac - 0.5);
			b = 0.0;
		}

		glColor4f(r, g, b, measurements[i].alpha);
		glVertex2f(points[measurements[i].p1].x, points[measurements[i].p1].y);
		glVertex2f(points[measurements[i].p2].x, points[measurements[i].p2].y);
	}
	glEnd();

	win.flip();
}

void randomize(double Z)
{
	for (int i = 2; i < points.size(); ++i) {
		points[i].x += Z * (rand() / (RAND_MAX + 1.0) - .5);
		points[i].y += Z * (rand() / (RAND_MAX + 1.0) - .5);
	}
}

void destress()
{
	double max_stress = 0.0;
	int p1, p2;

	for (int i = 0; i < measurements.size(); ++i) {
		if (measurements[i].p1 <= 1 ||
		    measurements[i].p2 <= 1)
		    	continue;
		
		if (measurements[i].stress > max_stress) {
			max_stress = measurements[i].stress;
			p1 = measurements[i].p1;
			p2 = measurements[i].p2;
		}
	}
	
	std::swap(points[p1], points[p2]);
}

int main(void)
{	
	GLWindow glw("foo", 640, 480, 32, false, 16, -1);
	int one = 1;

	srand((time_t)time(NULL));

	ioctl(0, FIONBIO, &one);
	reload();

	for ( ;; ) {
		char ch;
		while (read(0, &ch, 1) == 1) {
			if (ch == 'r') {
				reload();
			}
			if (ch == 'z') {
				randomize(0.005);
			}
			if (ch == 'Z') {
				randomize(0.1);
			}
			if (ch == 's') {
				destress();
			}
		}
		recalc();
		recalc();
		recalc();
		recalc();
		recalc();
		redraw(glw);
	}
}
